/* objmenu.c
 *
 * Copyright (C) 1999 vince (Vincent Saugey), MandrakeSoft <vince@mandrakesoft.com>
 * Copyright (C) 2002, 2003, 2004 Frederic Crozat, MandrakeSoft <fcrozat@mandrakesoft.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <string.h>
#include <strings.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <regex.h> 
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <langinfo.h>
#include "objmenu.h"
#include "callbacks.h"

#define BUF_SIZE 100000

GSList * attributes_list = NULL;

GSList * translation_list;

/* 
 * The two next function permit to create a list of attributes for the
 * menu entry manipulation 
 */

GSList *
get_attributes_list() 
{
  return attributes_list;
}

void
add_attributes_to_list(const gchar * attribute) {
  if (!g_slist_find_custom(attributes_list, (gpointer) attribute, (GCompareFunc)strcmp)) {
    attributes_list = g_slist_prepend (attributes_list, g_strdup(attribute));
  }
}

enum translation_state { 
	SEARCH,
	TRANSLATE,
	SUBTRANSLATE,
	SUBSTITUTE
};

struct translation_key {
	enum translation_state state;
	char * from;
	char * to;
	char * old_value;
	char * new_value;
};

void translation_list_free (gpointer *data, gpointer user_data) 
{
  g_return_if_fail (data != NULL);
  struct translation_key * translation_value = (struct translation_key *) data;

  g_free (translation_value->from);
  g_free (translation_value->to);
}

void load_translation_map () {
 FILE * translation_map; 
 char *file;
 char buf[BUF_SIZE];
 enum translation_state state = SEARCH;
 char *p;
 GString *new_value, *old_value;
 gboolean escaping, search_new_value;
 struct translation_key * translation_value;
 char *from, *to;
 char * orig;

 if (check_simplified())
     file = "/etc/menu-methods/simplified/translate_menus";
  else 
     file = "/etc/menu-methods/translate_menus";
 
 translation_map = fopen (file, "r");
 if (translation_map == NULL)
	 return;

 g_slist_foreach (translation_list, (GFunc) translation_list_free, NULL);
 g_slist_free (translation_list);

 translation_list = NULL;
 
 from = to = NULL;

 while (fgets (buf, BUF_SIZE, translation_map)) {

	 if (g_str_has_prefix (buf, "#"))
		continue;

	 switch (state) {
		 case SEARCH:
			orig = buf;
	 		if (g_str_has_prefix (buf, "substitute")) {
				state = SUBSTITUTE;
				orig = buf + 10;
			} 
			else if (g_str_has_prefix (buf, "translate")) {
				state = TRANSLATE;
				orig = buf + 9;
			} 
			else if (g_str_has_prefix (buf, "subtranslate")) {
				state = SUBTRANSLATE;
				orig = buf + 12;
		 	}
			p = strstr (orig, "->");
			if (p == NULL) 
				continue;
			from = g_strndup (orig, p-orig);
			g_strstrip (from);
			to= g_strdup (p+2);
			g_strstrip (to);

			break;

		 case TRANSLATE:
		 case SUBTRANSLATE:
		 case SUBSTITUTE:
			if (g_str_has_prefix (buf, "endtranslate")) {
		 		state = SEARCH;
				g_free (from);
				g_free (to);
				continue;
		 	}

			g_strstrip(buf);
	
			if (strlen(buf) == 0) 
				continue;

			p = buf;
			escaping = FALSE;
			search_new_value = FALSE;
			old_value = g_string_new ("");
			new_value = g_string_new ("");

	while (p && *p && *p != '\n') {
		switch (*p) {
			case '"':
				if (escaping) {
					escaping = FALSE;
					if (search_new_value)
						break;
					else 
						search_new_value = TRUE;
				} else {
					escaping = TRUE;
				}
				
				break;
			case ' ':
			case '\t':
				if (escaping)
					if (search_new_value) 
						g_string_append_c (new_value, *p);
					else 
						g_string_append_c (old_value, *p);
				else
					search_new_value = TRUE;
				break;
			default:
				if (search_new_value)
					g_string_append_c (new_value, *p);
				else {
					g_string_append_c (old_value, *p);
				}
				break;
		}
		p++;
	}

	translation_value = g_new0 (struct translation_key, 1);

	translation_value->from = g_strdup (from);
	translation_value->to = g_strdup (to);
	translation_value->old_value = old_value->str;
	translation_value->new_value = new_value->str;
	translation_value->state = state;

	translation_list = g_slist_append (translation_list, translation_value);

	g_string_free (old_value, FALSE);
	g_string_free (new_value, FALSE);

	 }
 }
 fclose(translation_map);
}

static
void apply_translation_map (gpointer data, gpointer user_data)
{
	struct translation_key * translation_value = data;
	GData ** entry = user_data;
	char * from, *to;

	from = g_datalist_get_data (entry, translation_value->from);
	if (from == NULL)
		return;
	switch (translation_value->state) {
		case TRANSLATE:
			if (strcmp (from,translation_value->old_value) != 0)
				return;
			g_datalist_set_data_full (entry, translation_value->to, g_strdup (translation_value->new_value), g_free);
		break;
			
		case SUBTRANSLATE:
			if (g_str_has_prefix (from, translation_value->old_value) != 0)
				return;
			g_datalist_set_data_full (entry, translation_value->to, g_strdup (translation_value->new_value), g_free);
			break;
		case SUBSTITUTE:
			if (g_str_has_prefix (from, translation_value->old_value) == FALSE)
				return;
			to = g_strconcat (translation_value->new_value, from+strlen(translation_value->old_value), NULL);
			g_datalist_set_data_full (entry, translation_value->from,
				       to, g_free);	
			break;
			
		default:
		break;
	}
}


static 
void canonize_entry(GData** obj) 
{
  gchar * pt, * section; 
  
  
  pt = section = SECTION(obj);
  while (section[0]=='/') { /* remove initial / */
    gchar *new_name = g_strdup(section + 1);
    REMOVE(obj, "section");
    g_datalist_set_data_full(obj, "section", new_name, g_free);
    section = new_name;
  }
 /* correction for double //  */
  for (pt = section ; *pt ; pt++) {
    if ((*pt == '/') && (*(pt+1) == '/')) {
      *(pt+1)='\0';
      memmove (section, (pt+2), strlen(pt+2));
    }
  }
 /* correction for not / end */
  if (*(--pt) != '/') {
    gchar *new_name = g_strdup_printf("%s/", section);
    REMOVE(obj, "section");
    g_datalist_set_data_full(obj, "section", new_name, g_free);
  }
  
  /* apply translation map */
  g_slist_foreach ( translation_list, apply_translation_map, obj);
 }



static 
int read_by_entry(char * buf,int maxsize, FILE * file) 
{
  gchar *str=buf;
  char * new_str;
  char * cleaned_str;
  buf[0]='\0';
  while (( new_str = fgets (str, maxsize - strlen(buf), file))) {
    str += strlen (str) - 1;
    while ((*str == ' ') || (*str == '\n')) str --;
    if (*str != '\\') {
      break;
    }
    *str=' ';
    str++;
  }
  cleaned_str = g_strstrip(buf);
  g_strlcpy (buf,cleaned_str,maxsize);
  return (int)new_str;
}


static 
char * parse_blank(char * parser) 
{
  while (*parser ==' ' || *parser == '\t') parser ++; 
  return parser;
}

static
char * get_token (char ** string, const char * delim) 
{
  *string  = parse_blank (*string);
  return strsep(string, delim);
}


static
GSList * read_list_menu_entry (FILE * menu_file) 
{
  GSList * list_menu_entry = NULL;
  gchar buf [BUF_SIZE];
  gchar *attribute, *value=NULL, * parser;
  gchar *newvalue;
  gchar *iterator;
  gchar *tmp;
  gboolean attribute_is_valid;
  GPatternSpec *pattern;

  g_return_val_if_fail (menu_file != NULL, NULL);
  pattern = g_pattern_spec_new("?package(*)");
  /* Read one entry by one entry */
  while (read_by_entry(buf, BUF_SIZE, menu_file)) {
    GData ** entry = g_new (GData*, 1);
    g_datalist_init (entry);
/*        g_print("Read %s", buf);*/
    parser = buf;
    attribute = get_token(&parser, ":#\n");
    if (!g_pattern_match_string (pattern, attribute)) {
      continue;
    }
    g_datalist_set_data_full(entry, "package", g_strdup(attribute), g_free);
    while (attribute && *attribute &&  parser && (*parser!='\n')) {
      if (*attribute == '#') {
	break;
      }
      attribute = get_token(&parser, "=#" );
      attribute_is_valid = TRUE;
      iterator = attribute;
      while (attribute_is_valid && *iterator) {
	attribute_is_valid &= g_ascii_isalpha(*iterator) || *iterator=='-' || *iterator=='_';
	if (!attribute_is_valid) {
	  g_print("attribute |%s| in \"%s\" is invalid\n",attribute, buf);
	}
	iterator++;
      }
      if (!attribute_is_valid) {
	break;
      }

      if (attribute && (strlen(attribute)> 3)) {
/*      g_print("attribute %s %s\n",attribute,parser);*/
	parser = parse_blank (parser); 
	if (*parser == '\"') {
	  int i=-2;
	  parser++;
	  value = strsep(&parser, "\"\n");
	  while (parser && parser[i] == '\\' && parser[i-1] != '\\') {
	    parser[i--]='\"';
	    strcat (value, strsep(&parser, "\"\n"));
	  }
	}
	else {
	  value = strsep(&parser, " \n" );
	}
	add_attributes_to_list(attribute);
/*	g_print("attribute: %s; value: %s\n", attribute, value);*/
	tmp = g_strcompress (value);
	if (g_utf8_validate (tmp, -1, NULL)) {
	  newvalue = tmp;
	}
	else {
	  newvalue = g_locale_to_utf8(g_strcompress(tmp), -1, NULL, NULL, NULL);
	  g_free (tmp);
	}
	g_datalist_set_data_full(entry, attribute, newvalue, g_free);
      }
    }
    if (TITLE(entry) && SECTION(entry) && NEEDS(entry)) {
      list_menu_entry = g_slist_prepend(list_menu_entry, entry);
    }
  }
  /* correction section for the obj list */
  g_slist_foreach (list_menu_entry, (GFunc)canonize_entry, NULL);
  return list_menu_entry;
}

static GSList * 
get_dir_file (const gchar * directory) 
{
  GSList * files=NULL;
  DIR * direc;
  struct dirent * file;
  struct stat file_stat;
  gchar * file_name;
  if ((direc = opendir(directory)) == NULL) {
    /*g_message(_("Cannot open directory: %s"), directory);*/
    return NULL;
  }
  while ((file = readdir(direc)) != NULL) {
    file_name = g_build_filename (directory, file->d_name, NULL);
    if (stat(file_name, &file_stat) == -1) {
      g_warning(_("Cannot get stat for file %s"), file_name);
      if (closedir(direc) == -1) {
	g_warning(_("Cannot close dir: %s"), directory);
      }
      g_free (file_name);
      return NULL;
    }
    if (S_ISREG(file_stat.st_mode) || S_ISLNK(file_stat.st_mode)) {
      files = g_slist_prepend(files, file_name);
    }
    else {
	    g_free (file_name);
    }
  }
  if (closedir(direc) == -1) {
    g_warning(_("Cannot close dir: %s"), directory);
    return NULL;
  }
  return files;
}

static int
cmp_basename(gchar * file1, gchar* file2) 
{
  gchar *path1, *path2;
  int retval;

  path1 = g_path_get_basename (file1);
  path2 = g_path_get_basename (file2);
  retval = strcmp (path1, path2);
  g_free (path1);
  g_free (path2);

  return retval;
}

static int
get_min(char **name, int name_len)
{
  int min = 0, i;

  for (i=0; i < name_len ; i++) {
	  if (name[i]) {
		  min = i;
		  break;
	  }
  }

  if (min < name_len -1) {
    for (i=min + 1; i < name_len ; i++) {
  	  if (name[i]) {
  		  if (strcmp(name[i], name[min]) < 0) {
			  min = i;
		  }
	  }
	  
    }
  }
  return min;
}

static 
GSList * get_menu_files (gboolean system) {
  GSList * tablFiles [4], * parser [4],* return_list = NULL;
  int i;
  if (system) {
    tablFiles[0]=NULL;
  }
  else {
    char * user_menu_dir;
    if (check_simplified()) {
      user_menu_dir = g_build_filename (g_get_home_dir(), ".menu-simplified",NULL);
    }
    else {
      user_menu_dir = g_build_filename (g_get_home_dir(), ".menu",NULL);
    }

    tablFiles[0] = get_dir_file (user_menu_dir);
    g_free(user_menu_dir);
  }

  if (check_simplified()) {
    tablFiles[1] = get_dir_file (SIMPLIFIED_MENU_OVERLOAD);
    tablFiles[2] = get_dir_file (SIMPLIFIED_SYSTEM_MENU_LIB);
    tablFiles[3] = get_dir_file (SYSTEM_MENU_LIB);
  }
  else {
    tablFiles[1] = get_dir_file (SYSTEM_MENU_OVERLOAD);
    tablFiles[2] = get_dir_file (SYSTEM_MENU_LIB);
    tablFiles[3] = NULL;
  }
  for (i=0; i<G_N_ELEMENTS(tablFiles);i++) {
    if (tablFiles[i]) {
      tablFiles[i] = g_slist_sort (tablFiles[i], (GCompareFunc) cmp_basename);
    }
  }
  for (i=0; i<G_N_ELEMENTS(parser);i++) {
    parser[i] = tablFiles[i];
  }
  while (parser[0] || parser[1] || parser[2] || parser[3]) {
    int min;
    char * cur_file;
    char * paths[G_N_ELEMENTS(parser)];
    for (i=0; i<G_N_ELEMENTS(parser); i++) {
      if (parser[i]) 
	paths[i] = g_path_get_basename((char*)parser[i]->data);
      else
	paths[i] = NULL;
    }    
    min = get_min(paths, G_N_ELEMENTS(paths));
    for (i=0; i<G_N_ELEMENTS(paths); i++) 
      g_free (paths[i]);

    if (parser[min]) {
      cur_file = g_strdup(parser[min]->data);
      return_list = g_slist_prepend(return_list, cur_file); 
      for (i=0; i<G_N_ELEMENTS(parser);i++) 
        if ((parser[i]) && (!cmp_basename((char*)parser[i]->data, (char *) cur_file))) {
  	  parser[i] = g_slist_next(parser[i]);
        }
    }
  }
  /* free memory */
  for (i=0; i<G_N_ELEMENTS(tablFiles);i++) {
    g_slist_foreach (tablFiles[i], (GFunc) g_free, NULL);
    if (tablFiles[i]) {
      g_slist_free(tablFiles[i]);
    }
  }
  return return_list;
}


static void 
concat_all_in_file (char * file, GSList * * list) 
{
    FILE * menu_file = fopen (file, "r");
    if (menu_file) {
      *list = g_slist_concat(*list, read_list_menu_entry (menu_file));
      fclose(menu_file);
    }
    else {
      g_print(_("Unable to open file %s\n"),file);
    }
}

GSList *
create_menu(gboolean system) {
  GSList * lFiles, * lEntries = NULL;

  lFiles = get_menu_files(system);
  g_slist_foreach (lFiles, (GFunc)concat_all_in_file , &lEntries);
  g_slist_foreach (lFiles, (GFunc) g_free, NULL);
  g_slist_free(lFiles);
  return lEntries;
}

GSList *
create_lib_menu()
{
  GSList * lFiles, * lEntries = NULL;
  if (check_simplified()) {
    lFiles = get_dir_file (SIMPLIFIED_SYSTEM_MENU_LIB);
    g_slist_foreach (lFiles, (GFunc)concat_all_in_file , &lEntries);
    g_slist_foreach (lFiles, (GFunc)g_free, NULL);
    g_slist_free(lFiles);
  }
  lFiles = get_dir_file (SYSTEM_MENU_LIB);
  g_slist_foreach (lFiles, (GFunc)concat_all_in_file , &lEntries);
  g_slist_foreach (lFiles, (GFunc)g_free, NULL);
  g_slist_free(lFiles);

  return lEntries;
}

static 
int cmp_attribute (char * attribute, GData ** obj1, GData ** obj2) {
  gchar * string1 = GET(obj1, attribute);
  gchar * string2 = GET(obj2, attribute);
  if (!string1) {
    return !string2;
  }
  return !string2 || !strcasecmp(string1, string2);
}


static 
gboolean is_hidden_entry (GData **obj) {
	return (((SECTION(obj) != NULL) && g_str_has_suffix (SECTION(obj), ".hidden/")) || strcmp (TITLE(obj), ".hidden") == 0);
}

static 
int cmp_all(GData ** obj1, GData ** obj2) {
  GSList *parser = attributes_list;
  for (parser = attributes_list; parser; parser = g_slist_next (parser)) {
    if (!cmp_attribute(parser->data, obj1, obj2)) {
      return TRUE;
    }
  }
  return FALSE;
}


static 
int include_in (GSList * list_elem, GSList * list_all) {
  GSList * iterator = list_elem;
  while (iterator) {
    if (!is_hidden_entry (iterator->data)) {
    	if (!g_slist_find_custom (list_all, iterator->data, (GCompareFunc) cmp_all)) {
      		return FALSE;
	}
    }
    iterator = g_slist_next(iterator);
  }
  return TRUE;
}


static 
GSList * delete_of_list (GSList * list_all, GSList * list_elem) {
  GSList * parser = list_elem;
  for (parser = list_elem; parser; parser = g_slist_next(parser)) {
    GSList * link = g_slist_find_custom (list_all, parser->data, (GCompareFunc) cmp_all);
    if (link)
      list_all = g_slist_remove_link(list_all , link);
  }
  return list_all;
}


/* copy of glib g_strescape which is UTF-8 friendly */
static 
char *
my_strescape (const gchar *source,
	     const gchar *exceptions)
{
  const guchar *p;
  gchar *dest;
  gchar *q;
  guchar excmap[256];
  
  g_return_val_if_fail (source != NULL, NULL);

  p = (guchar *) source;
  /* Each source byte needs maximally four destination chars (\777) */
  q = dest = g_malloc (strlen (source) * 4 + 1);

  memset (excmap, 0, 256);
  if (exceptions)
    {
      guchar *e = (guchar *) exceptions;

      while (*e)
	{
	  excmap[*e] = 1;
	  e++;
	}
    }

  while (*p)
    {
      if (excmap[*p])
	*q++ = *p;
      else
	{
	  switch (*p)
	    {
	    case '\b':
	      *q++ = '\\';
	      *q++ = 'b';
	      break;
	    case '\f':
	      *q++ = '\\';
	      *q++ = 'f';
	      break;
	    case '\n':
	      *q++ = '\\';
	      *q++ = 'n';
	      break;
	    case '\r':
	      *q++ = '\\';
	      *q++ = 'r';
	      break;
	    case '\t':
	      *q++ = '\\';
	      *q++ = 't';
	      break;
	    case '\\':
	      *q++ = '\\';
	      *q++ = '\\';
	      break;
	    case '"':
	      *q++ = '\\';
	      *q++ = '"';
	      break;
	    default:
		*q++ = *p;
	      break;
	    }
	}
      p++;
    }
  *q = 0;
  return dest;
}


static 
void write_attribute(FILE* file_menu, GData** obj, gchar * attribute) {
  gchar * value = GET(obj, attribute);
  if (value) {
    char *tmp;

    tmp = my_strescape (value, NULL);
    if (0==fprintf(file_menu, " %s=\"%s\"", attribute, tmp)) {
      g_error (_("cannot write menu to file"));
    }
    g_free (tmp);
  }
}


/* TODO  get the package who contains the executable */
static 
void write_menu(GData** obj, FILE * file_menu) {
  GSList * attributes;
  if (!file_menu)
    g_error (_("cannot write menu entry to the menu edit file"));

  if (GET(obj, "package")) {
    if (0==fprintf(file_menu, "%s:", (gchar *)GET(obj, "package")))
      g_error (_("cannot write menu entry to the menu edit file"));
  }
  else {
    if (0==fprintf(file_menu, "?package(menu):"))
      g_error (_("cannot write menu entry to the menu edit file"));
  }
  attributes = attributes_list;
  while (attributes) {
    if (strcmp("package", attributes->data)) {
      write_attribute(file_menu, obj, attributes->data);
    }
    attributes = g_slist_next(attributes);
  }
  fprintf(file_menu, "\n");
  return;
}


/* Verify the directory for the user menu file existe */
static 
void clean_directory(const gchar* dirname) {
  if (access(dirname, R_OK|W_OK) == -1) {
    if (mkdir (dirname, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
      g_error (_("cannot create directory:%s"), dirname);
    }
  }
  else {
    char * command = g_strdup_printf("/bin/rm -rf `/bin/ls %s/* 2>/dev/null | /bin/grep -v mdk_customization | /bin/grep -v enable_simplified | /bin/grep -v do-not-create-menu-link`", dirname);
    system(command);
    g_free(command);
  }
}


static
void force_charset_att(GData ** obj, char * charset){
  g_datalist_set_data_full(obj, "charset", g_strdup (charset), g_free);
}

/*
 * This next function save the list of entry in the menu directory
 * the list is !!! free !!! in this function but gtkobject of the menu 
 * entry don't destroy
 * First Argument is the list of menu
 * Second is the list of sources files to overload
 * 3 the destination dir
 * 4 the name of file where menudrake will be adding entry
 */

static
int save_lib (GSList *menus, GSList* srcfiles, gchar *destdir, gchar *menudrakefilename) {
  FILE * menudrakefile;
  gchar *name;
  gchar *path;

  clean_directory(destdir);
  for (; srcfiles; srcfiles = g_slist_next(srcfiles)) {
    FILE * curfile;
    gchar *srcfile, *destfile;
    GSList * srcmenus;
    srcfile = srcfiles->data;
    if (!(curfile = fopen (srcfile, "r"))) { 
      g_warning (_("cannot read menu file:%s"), srcfile); continue; 
    }
    srcmenus = read_list_menu_entry (curfile);
    fclose(curfile);    
    path =  g_path_get_basename(srcfile);
    destfile = g_build_filename(destdir,path, NULL);
    g_free (path);
    if (!include_in(srcmenus, menus)) { /* if not found all entries in list */
      /* create the overload user's file */
      FILE * file;
      GSList * menuparser;
      if (!(file = fopen(destfile, "w")))
	g_error (_("Cannot create file: %s"), destfile);
      /* comment system menu entry in user file */
      g_free(destfile);
      for (menuparser = srcmenus; menuparser; menuparser = g_slist_next(menuparser)) {
	GSList * elem = g_slist_find_custom(menus, menuparser->data, (GCompareFunc) cmp_all);

	if ((!elem) && (fprintf(file, "#") != 1))
	  g_error (_("cannot write menu entry to the menu edit file"));
	write_menu(menuparser->data, file);
      }
      fclose(file);
    }
    /*else
      if (((access(destfile, F_OK) == 0) && (unlink(destfile) == -1)))
      g_warning (_("cannot delete file %s"), destfile); */
    menus = delete_of_list(menus, srcmenus);
    g_slist_foreach(srcmenus, (GFunc) g_datalist_clear, NULL);
    g_slist_free(srcmenus);
  }

  /* writing the menu entry don't found in system menu */
  /* open the a file for the menu entry will be adding by the programme or die */
  name = g_build_filename (destdir, menudrakefilename, NULL);
  if (! (menudrakefile = fopen (name,"w")) ) {
    g_error (_("couldn't open the user menu file \"%s\" for writing"), menudrakefilename);
  }

  add_attributes_to_list("charset");
  g_slist_foreach(menus, (GFunc) force_charset_att, "utf8");
  
  g_slist_foreach(menus, (GFunc) write_menu, menudrakefile);
  fclose(menudrakefile);

  return TRUE;
}


int
save_user_menu(GSList * menus)
{
  gchar *destdir;
  GSList *srcfiles;
  int retcode;
  srcfiles = get_menu_files (TRUE);
  if (check_simplified()) {
    destdir = g_build_filename (g_get_home_dir(), ".menu-simplified",NULL);
  }
  else {
    destdir = g_build_filename (g_get_home_dir(),".menu",NULL);
  }
  retcode = save_lib(menus, srcfiles, destdir, "added_by_menudrake");
  /* free memory */
  g_slist_foreach (srcfiles, (GFunc) g_free, NULL);
  g_slist_free(srcfiles);
  g_slist_free(menus);
  return retcode;
}


int
save_system_menu(GSList * menus)
{
  gchar *destdir;
  GSList *srcfiles;
  int retcode;
  srcfiles = get_dir_file (SYSTEM_MENU_LIB);
  if (check_simplified()) {
    destdir = SIMPLIFIED_MENU_OVERLOAD;
  }
  else {
    destdir = SYSTEM_MENU_OVERLOAD;
  }
  retcode = save_lib(menus, srcfiles, destdir, "menudrakeentry");
  /* free memory */
  g_slist_foreach (srcfiles, (GFunc) g_free, NULL);
  g_slist_free(srcfiles);
  g_slist_free(menus);
  return retcode;
}
